#include <windows.h>

#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <assert.h>
#include <stdio.h>

#pragma pack(1)

typedef struct {
   int bpp;
   long width, height, width_bytes;
   char *bits;
} IMAGE;

typedef struct {
   BYTE idLen;
   BYTE ColorMapType;
   BYTE ImageTypeCode;
   WORD ColorMapOrigin;
   WORD ColorMapLength;
   BYTE ColorMapEntrySize;
   WORD XOrigin;
   WORD YOrigin;
   WORD Width;
   WORD Height;
   BYTE PixelSize;
   BYTE ImageDescriptor;
} TGAHEADER;

#pragma pack()

static int LoadTGA(const char *name, IMAGE *image)
{
   int tga_file, i;
   TGAHEADER tgah;
   DWORD dwSize;
   char *p;

   tga_file = _open(name, O_RDONLY|O_BINARY);
   if(tga_file == -1)
      return 0;

   if(_read(tga_file, &tgah, sizeof(TGAHEADER)) != sizeof(TGAHEADER)) {
      _close(tga_file);
      return 0;
   }

   if(tgah.ColorMapType != 0 || tgah.ImageTypeCode != 2 || tgah.PixelSize != 32) {
      _close(tga_file);
      return 0;
   }

   image->bpp = 32;
   image->width = tgah.Width;
   image->width_bytes = tgah.Width * (tgah.PixelSize >> 3);
   image->height = tgah.Height;

   dwSize = image->height * image->width_bytes;
   image->bits = malloc(dwSize);
   if(!image->bits) {
      _close(tga_file);
      return 0;
   }

   _lseek(tga_file, tgah.idLen, SEEK_CUR);

   for(i = tgah.Height - 1; i >= 0; i--) {
      p = image->bits + i * tgah.Width * 4;
      if(_read(tga_file, p, tgah.Width * 4) != tgah.Width * 4) {
         free(image->bits);
         _close(tga_file);
         return 0;
      }
   }
/*
   printf("idLen = %d\n"
          "ColorMapType = %d\n"
          "ImageTypeCode = %d\n"
          "ColorMapOrigin = %d\n"
          "ColorMapLength = %d\n"
          "ColorMapEntrySize = %d\n"
          "XOrigin = %d\n"
          "YOrigin = %d\n"
          "Width = %d\n"
          "Height = %d\n"
          "PixelSize = %d\n"
          "ImageDescriptor = %d\n",
          (int)tgah.idLen,
          (int)tgah.ColorMapType,
          (int)tgah.ImageTypeCode,
          (int)tgah.ColorMapOrigin,
          (int)tgah.ColorMapLength,
          (int)tgah.ColorMapEntrySize,
          (int)tgah.XOrigin,
          (int)tgah.YOrigin,
          (int)tgah.Width,
          (int)tgah.Height,
          (int)tgah.PixelSize,
          (int)tgah.ImageDescriptor);
*/
   _close(tga_file);

   return 1;
}


static void EMU_ConvertTo24(IMAGE *image, BYTE *bits24, int width_bytes)
{
    BYTE *s, *d;
    int i, j;

    assert(image->bpp == 32);

    for(i = 0; i < image->height; i++)
    {
	s = image->bits + (image->height - (i + 1)) * image->width_bytes;
	d = bits24 + i * width_bytes;

	for(j = 0; j < image->width; j++)
	{
	    *d++ = *s++;
	    *d++ = *s++;
	    *d++ = *s++;
	    s++;
	}
    }
}

/*
static void dump_dc_state(HDC hdc)
{
HBITMAP hbmp;
BITMAP bm;

printf("============\nhdc %p\n", hdc);
printf("BITSPIXEL %d\n", GetDeviceCaps(hdc, BITSPIXEL));
printf("HORZRES %d\n", GetDeviceCaps(hdc, HORZRES));
printf("VERTRES %d\n", GetDeviceCaps(hdc, VERTRES));
printf("LOGPIXELSX %d\n", GetDeviceCaps(hdc, LOGPIXELSX));
printf("LOGPIXELSY %d\n", GetDeviceCaps(hdc, LOGPIXELSY));
printf("GetBkColor 0x%08x\n", GetBkColor(hdc));
printf("GetTextColor 0x%08x\n", GetTextColor(hdc));
printf("GetBkMode %d\n", GetBkMode(hdc));

hbmp = CreateCompatibleBitmap(hdc, 1, 1);
assert(GetObject(hbmp, sizeof(BITMAP), &bm) == sizeof(BITMAP));
printf("Bit depth of compatible bitmap %d\n", bm.bmBitsPixel);
DeleteObject(hbmp);
}
*/

static BOOL ShowImage32(IMAGE *image32, HWND hWnd)
{
    BITMAPINFO bmi;
    BITMAP bm;
    HBITMAP hBmp, hBmpOld;
//    LOGFONT lf;
//    HFONT hFont, hFontOld;
    HDC hWindowDC, hMemDC;
    BYTE *bits24;
//    RECT rc;
//    DIBSECTION ds;

    assert(image32->bpp == 32);

/*
    lf.lfCharSet = ANSI_CHARSET;
    lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    lf.lfEscapement = 0;
    strcpy(lf.lfFaceName, "Arial");
    lf.lfHeight = -2048;
    lf.lfItalic = 0;
    lf.lfOrientation = 0;
    lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
    lf.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
    lf.lfQuality = DEFAULT_QUALITY;
    lf.lfStrikeOut = FALSE;
    lf.lfUnderline = FALSE;
    lf.lfWeight = FW_NORMAL;
    lf.lfWidth = 0;
    hFont = CreateFontIndirect(&lf);
*/

    hWindowDC = GetDC(hWnd);
    assert(hWindowDC);

    memset(&bmi, 0, sizeof(BITMAPINFO));

    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = image32->width;
    bmi.bmiHeader.biHeight = image32->height;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 24;
    bmi.bmiHeader.biCompression = BI_RGB;
    bmi.bmiHeader.biSizeImage = 0;
    bmi.bmiHeader.biXPelsPerMeter = 0;
    bmi.bmiHeader.biYPelsPerMeter = 0;
    bmi.bmiHeader.biClrUsed = 0;
    bmi.bmiHeader.biClrImportant = 0;

    hBmp = CreateDIBSection(hWindowDC, &bmi, DIB_RGB_COLORS, &bits24, NULL, 0);
    assert(hBmp);
printf("DIB section handle %p\n", hBmp);

    assert(GetObject(hBmp, sizeof(BITMAP), &bm) == sizeof(BITMAP));
    EMU_ConvertTo24(image32, bits24, bm.bmWidthBytes);

//    SetBkColor(hWindowDC, RGB(0, 0xff, 0));
//    SetBkMode(hWindowDC, 1);
//    dump_dc_state(hWindowDC);
    hMemDC = CreateCompatibleDC(hWindowDC);
//    dump_dc_state(hMemDC);

    hBmpOld = SelectObject(hMemDC, hBmp);

//    GetPixel(hMemDC, 0, 0);

//printf("current selected bitmap %p\n", GetCurrentObject(hMemDC, OBJ_BITMAP));
//GetObject(hBmp, sizeof(DIBSECTION), &ds);
//printf("ds.dsBm.bmBits = %p\n", ds.dsBm.bmBits);

//    memset((char *)ds.dsBm.bmBits + ds.dsBm.bmWidthBytes*ds.dsBm.bmHeight/2, 0,
//    memset(ds.dsBm.bmBits, 0,
//	   ds.dsBm.bmWidthBytes*ds.dsBm.bmHeight/2);

#if 0
    GetObject(hBmp, sizeof(BITMAP), &bm);

    rc.left = 0;
    rc.top = 0;
    rc.right = bm.bmWidth;
    rc.bottom = bm.bmHeight;
    hFontOld = SelectObject(hMemDC, hFont);
    SetBkColor(hMemDC, RGB(0, 0xff, 0));
//    ExtTextOut(hMemDC, 0, 0, ETO_OPAQUE, &rc, NULL, 0, NULL);
    SelectObject(hMemDC, hFontOld);
#endif

    assert(BitBlt(hWindowDC, 0, 0, image32->width, image32->height, hMemDC, 0, 0, SRCCOPY));

    SelectObject(hMemDC, hBmpOld);
    DeleteDC(hMemDC);
    DeleteObject(hBmp);

    ReleaseDC(hWnd, hWindowDC);

    return TRUE;
}

static BOOL show_tga(HWND hwnd, const char *name)
{
    IMAGE image32;

    if (!LoadTGA(name, &image32)) return FALSE;
    return ShowImage32(&image32, hwnd);
}

static void fill_wnd(HWND hwnd)
{
    HDC hdc;
    HBRUSH hbrush1, hbrush2, hbrush_old;
    RECT rc;
    LONG height;

    GetClientRect(hwnd, &rc);
    height = (rc.bottom - rc.top) / 2;

    hdc = GetDC(hwnd);
    assert(hdc);
    SetROP2(hdc, R2_COPYPEN);

    hbrush1 = CreateSolidBrush(PALETTERGB(0xff,0,0));
    hbrush2 = CreateSolidBrush(PALETTERGB(0,0xff,0));
    hbrush_old = SelectObject(hdc, hbrush1);
//    FillRect(hdc, &rc, hbrush);
    PatBlt(hdc, rc.left, rc.top, rc.right, rc.top + height, PATCOPY);

    SelectObject(hdc, hbrush2);
    Rectangle(hdc, rc.left, rc.top + height, rc.right, rc.top + height*2);

    SelectObject(hdc, hbrush_old);
    ReleaseDC(hwnd, hdc);

    DeleteObject(hbrush1);
    DeleteObject(hbrush2);
}

BOOL CALLBACK MainDlgProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch(Msg) {
    case WM_CLOSE:
	EndDialog(hDlg, 0);
	break;

    case WM_PAINT: {
	PAINTSTRUCT ps;
	HDC hDC;

	hDC = BeginPaint(hDlg, &ps);
	EndPaint(hDlg, &ps);
	break;
    }

    case WM_COMMAND:
	switch(LOWORD(wParam)) {
	    case 0x555:
		show_tga(GetDlgItem(hDlg, 0x100), "sample.tga");
//		fill_wnd(GetDlgItem(hDlg, 0x100));
		return TRUE;

	    default: break;
	}
	break;
	
    default: break;
    }

    return FALSE;
}

int main(void)
{
    return DialogBox(GetModuleHandle(0), MAKEINTRESOURCE(1), NULL, MainDlgProc);
}
